unit DbxDataClxUnit;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs, FMTBcd,
  QStdCtrls, QDBCtrls, QExtCtrls, QGrids, QDBGrids, DB, DBClient, Provider,
  SqlExpr, QButtons, DBLogDlg, DbxDataClxRecError, DBXpress;

type
  TTestForm = class(TForm)
    SQLConn: TSQLConnection;
    SQLDs: TSQLDataSet;
    DataSetProvider1: TDataSetProvider;
    CDS: TClientDataSet;
    DataSource1: TDataSource;
    Panel1: TPanel;
    DriversListBox: TListBox;
    ConnectionsListBox: TListBox;
    TablesListBox: TListBox;
    FieldsListBox: TListBox;
    Panel2: TPanel;
    DBGrid1: TDBGrid;
    ControlBar1: TControlBar;
    DBNavigator1: TDBNavigator;
    BitBtn1: TBitBtn;
    Panel3: TPanel;
    DBImage1: TDBImage;
    Label1: TLabel;
    DBMemo1: TDBMemo;
    Label2: TLabel;
    procedure DriversListBoxDblClick(Sender: TObject);
    procedure ConnectionsListBoxDblClick(Sender: TObject);
    procedure TablesListBoxDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TablesListBoxClick(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure DriversListBoxClick(Sender: TObject);
    procedure ConnectionsListBoxClick(Sender: TObject);
    procedure CDSReconcileError(DataSet: TCustomClientDataSet;
      E: EReconcileError; UpdateKind: TUpdateKind;
      var Action: TReconcileAction);
  private
    procedure OpenTable;
    procedure CloseAll;
    function GetSelected(ListBox: TListBox): string;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  TestForm: TTestForm;

implementation

{$R *.xfm}
{ TTestForm }


procedure TTestForm.CloseAll;
begin
  BitBtn1.Enabled := False;
  Label1.Caption := '';
  Label2.Caption := '';
  DBImage1.DataField := '';
  DBMemo1.DataField := '';
  DBImage1.Visible := False;
  DBMemo1.Visible := False;
  CDS.Close;
  SQLDs.Close;
end;

procedure TTestForm.OpenTable;
var
  I: Integer;
  DataType: TFieldType;

begin
  try
    CloseAll;
    SqlDs.CommandText := GetSelected( TablesListBox );
    CDS.Active := True;
    BitBtn1.Enabled := True;
    for I := 0 to CDS.FieldCount - 1 do
    begin
      DataType := CDS.Fields[I].DataType;
      if DataType in [ftGraphic, ftBlob, ftMemo] then
      begin
        if DataType in [ftGraphic, ftBlob] then
        begin
          if DBImage1.DataField = '' then
          begin
            DBImage1.DataField := CDS.Fields[I].FieldName;
            DBImage1.Visible := True;
            Label1.Caption := DBImage1.DataField;
          end;
        end else
          if DBMemo1.DataField = '' then
          begin
            DBMemo1.DataField := CDS.Fields[I].FieldName;
            DBMemo1.Visible := True;
            Label2.Caption := DBMemo1.DataField;
          end;
      end;
    end;
  Except
    on e:Exception do
      ShowMessage(e.Message);
  end;
end;

function TTestForm.GetSelected(ListBox: TListBox): String;
var
  i : Integer;
begin
  Result := '';
  for i := 0 to ListBox.Items.Count -1 do
  begin
    if ListBox.Selected[i] then
    begin
      Result := ListBox.Items[I];
      break;
    end;
  end;
  if (ListBox.Items.Count > 0) and (Result = '') then Result := ListBox.Items[0];
end;

procedure TTestForm.DriversListBoxDblClick(Sender: TObject);
begin
  CloseAll;
  ConnectionsListBox.Items.Clear;
  TablesListBox.Items.Clear;
  FieldsListBox.Items.Clear;
  GetConnectionNames(ConnectionsListBox.Items, GetSelected(DriversListBox));
end;

procedure TTestForm.ConnectionsListBoxDblClick(Sender: TObject);
begin
  if ConnectionsListBox.Items.Count = 0 then exit;
  CloseAll;
  SQLConn.Connected := False;
  SQLConn.LoadParamsOnConnect := True;
  SQLConn.ConnectionName := GetSelected(ConnectionsListBox);
  SQLConn.Connected := True;
  TablesListBox.Items.Clear;
  FieldsListBox.Items.Clear;
  SQLConn.GetTableNames(TablesListBox.Items, False);
end;

procedure TTestForm.TablesListBoxDblClick(Sender: TObject);
begin
  if SQLConn.Connected then
    if TablesListBox.Items.Count > 0 then
      OpenTable;
end;

procedure TTestForm.FormCreate(Sender: TObject);
begin
  CloseAll;
  GetDriverNames(DriversListBox.Items);
end;

procedure TTestForm.TablesListBoxClick(Sender: TObject);
begin
  if SQLConn.Connected then
  begin
    FieldsListBox.Items.Clear;
    if TablesListBox.Items.Count > 0 then
      SqlConn.GetFieldNames(GetSelected(TablesListBox), FieldsListBox.Items);
  end;
end;

procedure TTestForm.BitBtn1Click(Sender: TObject);
begin
  if SQLConn.Connected then
    if CDS.Active then
      CDS.ApplyUpdates(-1);
end;

procedure TTestForm.DriversListBoxClick(Sender: TObject);
begin
  DriversListBoxDblClick(Sender);
end;

procedure TTestForm.ConnectionsListBoxClick(Sender: TObject);
begin
  ConnectionsListBoxDblClick(Sender);
end;

procedure TTestForm.CDSReconcileError(DataSet: TCustomClientDataSet;
  E: EReconcileError; UpdateKind: TUpdateKind;
  var Action: TReconcileAction);
begin
  Action := HandleReconcileError(DataSet, UpdateKind, E);
end;

end.
